@@ -1,7 +1,7 @@
 /*
  * This program source code file is part of KiCad, a free EDA CAD application.
  *
- * Copyright (C) 2004-2024 KiCad Developers.
+ * Copyright The KiCad Developers.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -25,10 +25,11 @@
 #include <math_for_graphics.h>
 #include <board_design_settings.h>
 #include <footprint.h>
+#include <layer_range.h>
 #include <pcb_shape.h>
 #include <pad.h>
 #include <pcb_track.h>
-#include <core/thread_pool.h>
+#include <thread_pool.h>
 #include <zone.h>
 
 #include <geometry/seg.h>
@@ -40,6 +41,7 @@
 #include <drc/drc_item.h>
 #include <drc/drc_rule.h>
 #include <drc/drc_test_provider_clearance_base.h>
+#include <drc/drc_creepage_utils.h>
 #include <pcb_dimension.h>
 
 #include <future>
@@ -90,12 +92,12 @@ private:
      * @param other item against which to test the track item
      * @return false if there is a clearance violation reported, true if there is none
      */
-    bool testSingleLayerItemAgainstItem( BOARD_CONNECTED_ITEM* item, SHAPE* itemShape,
-                                         PCB_LAYER_ID layer, BOARD_ITEM* other );
+    bool testSingleLayerItemAgainstItem( BOARD_ITEM* item, SHAPE* itemShape, PCB_LAYER_ID layer,
+                                         BOARD_ITEM* other );
 
     void testTrackClearances();
 
-    void testPadAgainstItem( PAD* pad, SHAPE* padShape, PCB_LAYER_ID layer, BOARD_ITEM* other );
+    bool testPadAgainstItem( PAD* pad, SHAPE* padShape, PCB_LAYER_ID layer, BOARD_ITEM* other );
 
     void testPadClearances();
 
@@ -113,7 +115,7 @@ private:
             : layers(), has_error( false ) {}
 
         checked( PCB_LAYER_ID aLayer )
-            : layers( aLayer ), has_error( false ) {}
+            : layers( { aLayer } ), has_error( false ) {}
 
         LSET layers;
         bool has_error;
@@ -196,7 +198,7 @@ bool DRC_TEST_PROVIDER_COPPER_CLEARANCE::Run()
 }
 
 
-bool DRC_TEST_PROVIDER_COPPER_CLEARANCE::testSingleLayerItemAgainstItem( BOARD_CONNECTED_ITEM* item,
+bool DRC_TEST_PROVIDER_COPPER_CLEARANCE::testSingleLayerItemAgainstItem( BOARD_ITEM* item,
                                                                          SHAPE* itemShape,
                                                                          PCB_LAYER_ID layer,
                                                                          BOARD_ITEM* other )
@@ -209,12 +211,19 @@ bool DRC_TEST_PROVIDER_COPPER_CLEARANCE::testSingleLayerItemAgainstItem( BOARD_C
     int            actual;
     VECTOR2I       pos;
     bool           has_error = false;
-    int            otherNet = 0;
+    NETINFO_ITEM*  net = nullptr;
+    NETINFO_ITEM*  otherNet = nullptr;
+
+    if( BOARD_CONNECTED_ITEM* connectedItem = dynamic_cast<BOARD_CONNECTED_ITEM*>( item ) )
+        net = connectedItem->GetNet();
+
+    NETINFO_ITEM*  trackNet = net;
 
     if( BOARD_CONNECTED_ITEM* connectedItem = dynamic_cast<BOARD_CONNECTED_ITEM*>( other ) )
-        otherNet = connectedItem->GetNetCode();
+        otherNet = connectedItem->GetNet();
 
-    std::shared_ptr<SHAPE> otherShape = other->GetEffectiveShape( layer );
+    std::shared_ptr<SHAPE> otherShapeStorage = other->GetEffectiveShape( layer );
+    SHAPE* otherShape = otherShapeStorage.get();
 
     if( other->Type() == PCB_PAD_T )
     {
@@ -232,6 +241,15 @@ bool DRC_TEST_PROVIDER_COPPER_CLEARANCE::testSingleLayerItemAgainstItem( BOARD_C
 
     if( constraint.GetSeverity() != RPT_SEVERITY_IGNORE && clearance > 0 )
     {
+        // Collide (and generate violations) based on a well-defined order so that exclusion
+        // checking against previously-generated violations will work.
+        if( item->m_Uuid > other->m_Uuid )
+        {
+            std::swap( item, other );
+            std::swap( itemShape, otherShape );
+            std::swap( net, otherNet );
+        }
+
         // Special processing for track:track intersections
         if( item->Type() == PCB_TRACE_T && other->Type() == PCB_TRACE_T )
         {
@@ -253,9 +271,9 @@ bool DRC_TEST_PROVIDER_COPPER_CLEARANCE::testSingleLayerItemAgainstItem( BOARD_C
             }
         }
 
-        if( itemShape->Collide( otherShape.get(), clearance - m_drcEpsilon, &actual, &pos ) )
+        if( itemShape->Collide( otherShape, clearance - m_drcEpsilon, &actual, &pos ) )
         {
-            if( m_drcEngine->IsNetTieExclusion( item->GetNetCode(), layer, pos, other ) )
+            if( trackNet && m_drcEngine->IsNetTieExclusion( trackNet->GetNetCode(), layer, pos, other ) )
             {
                 // Collision occurred as track was entering a pad marked as a net-tie.  We
                 // allow these.
@@ -265,8 +283,9 @@ bool DRC_TEST_PROVIDER_COPPER_CLEARANCE::testSingleLayerItemAgainstItem( BOARD_C
                 std::shared_ptr<DRC_ITEM> drce = DRC_ITEM::Create( DRCE_SHORTING_ITEMS );
                 wxString msg;
 
-                msg.Printf( _( "(nets %s and %s)" ), item->GetNetname(),
-                            static_cast<BOARD_CONNECTED_ITEM*>( other )->GetNetname() );
+                msg.Printf( _( "(nets %s and %s)" ),
+                            net ? net->GetNetname() : _( "<no net>" ),
+                            otherNet ? otherNet->GetNetname() : _( "<no net>" ) );
 
                 drce->SetErrorMessage( drce->GetErrorText() + wxS( " " ) + msg );
                 drce->SetItems( item, other );
@@ -289,7 +308,7 @@ bool DRC_TEST_PROVIDER_COPPER_CLEARANCE::testSingleLayerItemAgainstItem( BOARD_C
                 drce->SetItems( item, other );
                 drce->SetViolatingRule( constraint.GetParentRule() );
 
-                reportViolation( drce, pos, layer );
+                ReportAndShowPathCuToCu( drce, pos, layer, item, other, layer, actual );
                 has_error = true;
 
                 if( !m_drcEngine->GetReportAllTrackErrors() )
@@ -302,7 +321,7 @@ bool DRC_TEST_PROVIDER_COPPER_CLEARANCE::testSingleLayerItemAgainstItem( BOARD_C
     {
         std::array<BOARD_ITEM*, 2> a{ item, other };
         std::array<BOARD_ITEM*, 2> b{ other, item };
-        std::array<SHAPE*, 2>      a_shape{ itemShape, otherShape.get() };
+        std::array<SHAPE*, 2>      a_shape{ itemShape, otherShape };
 
         for( size_t ii = 0; ii < 2; ++ii )
         {
@@ -314,6 +333,7 @@ bool DRC_TEST_PROVIDER_COPPER_CLEARANCE::testSingleLayerItemAgainstItem( BOARD_C
             {
                 continue;
             }
+
             if( b[ii]->Type() == PCB_VIA_T )
             {
                 if( b[ii]->GetLayerSet().Contains( layer ) )
@@ -345,7 +365,7 @@ bool DRC_TEST_PROVIDER_COPPER_CLEARANCE::testSingleLayerItemAgainstItem( BOARD_C
                     drce->SetItems( a[ii], b[ii] );
                     drce->SetViolatingRule( constraint.GetParentRule() );
 
-                    reportViolation( drce, pos, layer );
+                    ReportAndShowPathCuToCu( drce, pos, layer, a[ii], b[ii], layer, actual );
                     return false;
                 }
             }
@@ -376,6 +396,43 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testItemAgainstZone( BOARD_ITEM* aItem,
     if( !worstCaseBBox.Intersects( aZone->GetBoundingBox() ) )
         return;
 
+    FOOTPRINT* parentFP = aItem->GetParentFootprint();
+
+    // Ignore graphic items which implement a net-tie to the zone's net on the layer being tested.
+    if( parentFP && parentFP->IsNetTie() && dynamic_cast<PCB_SHAPE*>( aItem ) )
+    {
+        std::set<PAD*> allowedNetTiePads;
+
+        for( PAD* pad : parentFP->Pads() )
+        {
+            if( pad->GetNetCode() == aZone->GetNetCode() && aZone->GetNetCode() != 0 )
+            {
+                if( pad->IsOnLayer( aLayer ) )
+                    allowedNetTiePads.insert( pad );
+
+                for( PAD* other : parentFP->GetNetTiePads( pad ) )
+                {
+                    if( other->IsOnLayer( aLayer ) )
+                        allowedNetTiePads.insert( other );
+                }
+            }
+        }
+
+        if( !allowedNetTiePads.empty() )
+        {
+            std::shared_ptr<SHAPE> itemShape = aItem->GetEffectiveShape();
+
+            for( PAD* pad : allowedNetTiePads )
+            {
+                if( pad->GetBoundingBox().Intersects( itemBBox )
+                        && pad->GetEffectiveShape( aLayer )->Collide( itemShape.get() ) )
+                {
+                    return;
+                }
+            }
+        }
+    }
+
     bool testClearance = !m_drcEngine->IsErrorLimitExceeded( DRCE_CLEARANCE );
     bool testHoles = !m_drcEngine->IsErrorLimitExceeded( DRCE_HOLE_CLEARANCE );
 
@@ -424,8 +481,7 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testItemAgainstZone( BOARD_ITEM* aItem,
             drce->SetErrorMessage( drce->GetErrorText() + wxS( " " ) + msg );
             drce->SetItems( aItem, aZone );
             drce->SetViolatingRule( constraint.GetParentRule() );
-
-            reportViolation( drce, pos, aLayer );
+            ReportAndShowPathCuToCu( drce, pos, aLayer, aItem, aZone, aLayer, actual );
         }
     }
 
@@ -463,8 +519,7 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testItemAgainstZone( BOARD_ITEM* aItem,
                     drce->SetErrorMessage( drce->GetErrorText() + wxS( " " ) + msg );
                     drce->SetItems( aItem, aZone );
                     drce->SetViolatingRule( constraint.GetParentRule() );
-
-                    reportViolation( drce, pos, aLayer );
+                    ReportAndShowPathCuToCu( drce, pos, aLayer, aItem, aZone, aLayer, actual );
                 }
             }
         }
@@ -548,8 +603,7 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testKnockoutTextAgainstZone( BOARD_ITEM
             drce->SetErrorMessage( drce->GetErrorText() + wxS( " " ) + msg );
             drce->SetItems( aText, aZone );
             drce->SetViolatingRule( constraint.GetParentRule() );
-
-            reportViolation( drce, pos, layer );
+            ReportAndShowPathCuToCu( drce, pos, layer, aText, aZone, layer, actual );
         }
     }
 }
@@ -598,8 +652,8 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testTrackClearances()
                             std::lock_guard<std::mutex> lock( checkedPairsMutex );
                             auto it = checkedPairs.find( { a, b } );
 
-                            if( it != checkedPairs.end() && ( it->second.layers.test( layer )
-                                    || ( it->second.has_error && !m_drcEngine->GetReportAllTrackErrors() ) ) )
+                            if( it != checkedPairs.end()
+                                    && ( it->second.layers.test( layer ) || ( it->second.has_error ) ) )
                             {
                                 return false;
                             }
@@ -634,25 +688,27 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testTrackClearances()
                                 }
                             }
 
-                            BOARD_ITEM* a = track;
-                            BOARD_ITEM* b = other;
-
-                            // store canonical order so we don't collide in both directions
-                            // (a:b and b:a)
-                            if( static_cast<void*>( a ) > static_cast<void*>( b ) )
-                                std::swap( a, b );
-
                             // If we get an error, mark the pair as having a clearance error already
                             if( !testSingleLayerItemAgainstItem( track, trackShape.get(), layer, other ) )
                             {
-                                std::lock_guard<std::mutex> lock( checkedPairsMutex );
-                                auto it = checkedPairs.find( { a, b } );
-
-                                if( it != checkedPairs.end() )
-                                    it->second.has_error = true;
-
                                 if( !m_drcEngine->GetReportAllTrackErrors() )
+                                {
+                                    BOARD_ITEM* a = track;
+                                    BOARD_ITEM* b = other;
+
+                                    // store canonical order so we don't collide in both directions
+                                    // (a:b and b:a)
+                                    if( static_cast<void*>( a ) > static_cast<void*>( b ) )
+                                        std::swap( a, b );
+
+                                    std::lock_guard<std::mutex> lock( checkedPairsMutex );
+                                    auto it = checkedPairs.find( { a, b } );
+
+                                    if( it != checkedPairs.end() )
+                                        it->second.has_error = true;
+
                                     return false;   // We're done with this track
+                                }
                             }
 
                             return !m_drcEngine->IsCancelled();
@@ -691,7 +747,7 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testTrackClearances()
 }
 
 
-void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testPadAgainstItem( PAD* pad, SHAPE* padShape,
+bool DRC_TEST_PROVIDER_COPPER_CLEARANCE::testPadAgainstItem( PAD* pad, SHAPE* padShape,
                                                              PCB_LAYER_ID aLayer,
                                                              BOARD_ITEM* other )
 {
@@ -765,13 +821,14 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testPadAgainstItem( PAD* pad, SHAPE* pa
     }
 
     if( !testClearance && !testShorting && !testHoles )
-        return;
+        return true;
 
     std::shared_ptr<SHAPE> otherShape = other->GetEffectiveShape( aLayer );
     DRC_CONSTRAINT         constraint;
     int                    clearance = 0;
     int                    actual = 0;
     VECTOR2I               pos;
+    bool                   has_error = false;
 
     if( otherPad && pad->SameLogicalPadAs( otherPad ) )
     {
@@ -781,12 +838,12 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testPadAgainstItem( PAD* pad, SHAPE* pa
         if( testShorting )
         {
             if( pad->GetNetCode() == 0 || pad->GetNetCode() == otherPad->GetNetCode() )
-                return;
+                return true;
 
             if( pad->GetShortNetname().StartsWith( wxS( "unconnected-(" ) )
                     && otherPad->GetShortNetname().StartsWith( wxS( "unconnected-(" ) ) )
             {
-                return;
+                return true;
             }
 
             std::shared_ptr<DRC_ITEM> drce = DRC_ITEM::Create( DRCE_SHORTING_ITEMS );
@@ -800,9 +857,10 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testPadAgainstItem( PAD* pad, SHAPE* pa
             drce->SetItems( pad, otherPad );
 
             reportViolation( drce, otherPad->GetPosition(), aLayer );
+            has_error = true;
         }
 
-        return;
+        return !has_error;
     }
 
     if( testClearance || testShorting )
@@ -820,7 +878,7 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testPadAgainstItem( PAD* pad, SHAPE* pa
                     // Pads connected to pads of a net-tie footprint are allowed to collide
                     // with the net-tie footprint's graphics.
                 }
-                else if( actual == 0 && otherNet && testShorting )
+                else if( actual == 0 && padNet && otherNet && testShorting )
                 {
                     std::shared_ptr<DRC_ITEM> drce = DRC_ITEM::Create( DRCE_SHORTING_ITEMS );
                     wxString msg;
@@ -833,6 +891,7 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testPadAgainstItem( PAD* pad, SHAPE* pa
                     drce->SetItems( pad, other );
 
                     reportViolation( drce, pos, aLayer );
+                    has_error = true;
                     testHoles = false;  // No need for multiple violations
                 }
                 else if( testClearance )
@@ -846,8 +905,8 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testPadAgainstItem( PAD* pad, SHAPE* pa
                     drce->SetErrorMessage( drce->GetErrorText() + wxS( " " ) + msg );
                     drce->SetItems( pad, other );
                     drce->SetViolatingRule( constraint.GetParentRule() );
-
-                    reportViolation( drce, pos, aLayer );
+                    ReportAndShowPathCuToCu( drce, pos, aLayer, pad, other, aLayer, actual );
+                    has_error = true;
                     testHoles = false;  // No need for multiple violations
                 }
             }
@@ -878,8 +937,8 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testPadAgainstItem( PAD* pad, SHAPE* pa
             drce->SetErrorMessage( drce->GetErrorText() + wxS( " " ) + msg );
             drce->SetItems( pad, other );
             drce->SetViolatingRule( constraint.GetParentRule() );
-
-            reportViolation( drce, pos, aLayer );
+            ReportAndShowPathCuToCu( drce, pos, aLayer, pad, other, aLayer, actual );
+            has_error = true;
             testHoles = false;  // No need for multiple violations
         }
     }
@@ -901,6 +960,7 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testPadAgainstItem( PAD* pad, SHAPE* pa
             drce->SetViolatingRule( constraint.GetParentRule() );
 
             reportViolation( drce, pos, aLayer );
+            has_error = true;
             testHoles = false;  // No need for multiple violations
         }
     }
@@ -920,10 +980,12 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testPadAgainstItem( PAD* pad, SHAPE* pa
             drce->SetErrorMessage( drce->GetErrorText() + wxS( " " ) + msg );
             drce->SetItems( pad, otherVia );
             drce->SetViolatingRule( constraint.GetParentRule() );
-
-            reportViolation( drce, pos, aLayer );
+            ReportAndShowPathCuToCu( drce, pos, aLayer, pad, otherVia, aLayer, actual );
+            has_error = true;
         }
     }
+
+    return !has_error;
 }
 
 
@@ -938,7 +1000,8 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testPadClearances( )
 
     reportAux( wxT( "Testing %d pads..." ), count );
 
-    std::unordered_map<PTR_PTR_CACHE_KEY, int> checkedPairs;
+    std::unordered_map<PTR_PTR_CACHE_KEY, layers_checked> checkedPairs;
+    std::mutex                                            checkedPairsMutex;
 
     LSET boardCopperLayers = LSET::AllCuMask( m_board->GetCopperLayerCount() );
 
@@ -968,20 +1031,34 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testPadClearances( )
                                         if( static_cast<void*>( a ) > static_cast<void*>( b ) )
                                             std::swap( a, b );
 
-                                        if( checkedPairs.find( { a, b } ) != checkedPairs.end() )
+                                        std::lock_guard<std::mutex> lock( checkedPairsMutex );
+                                        auto it = checkedPairs.find( { a, b } );
+
+                                        if( it != checkedPairs.end()
+                                                && ( it->second.layers.test( layer ) || it->second.has_error ) )
                                         {
                                             return false;
                                         }
                                         else
                                         {
-                                            checkedPairs[ { a, b } ] = 1;
+                                            checkedPairs[ { a, b } ].layers.set( layer );
                                             return true;
                                         }
                                     },
                                     // Visitor
                                     [&]( BOARD_ITEM* other ) -> bool
                                     {
-                                        testPadAgainstItem( pad, padShape.get(), layer, other );
+                                        if( !testPadAgainstItem( pad, padShape.get(), layer, other ) )
+                                        {
+                                            BOARD_ITEM* a = pad;
+                                            BOARD_ITEM* b = other;
+
+                                            std::lock_guard<std::mutex> lock( checkedPairsMutex );
+                                            auto it = checkedPairs.find( { a, b } );
+
+                                            if( it != checkedPairs.end() )
+                                                it->second.has_error = true;
+                                        }
 
                                         return !m_drcEngine->IsCancelled();
                                     },
@@ -1011,7 +1088,7 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testPadClearances( )
 }
 
 
-void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testGraphicClearances( )
+void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testGraphicClearances()
 {
     thread_pool&        tp = GetKiCadThreadPool();
     size_t              count = m_board->Drawings().size();
@@ -1055,6 +1132,7 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testGraphicClearances( )
             };
 
     std::unordered_map<PTR_PTR_CACHE_KEY, layers_checked> checkedPairs;
+    std::mutex                                            checkedPairsMutex;
 
     auto testCopperGraphic =
             [&]( PCB_SHAPE* aShape )
@@ -1085,6 +1163,7 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testGraphicClearances( )
                                 if( static_cast<void*>( a ) > static_cast<void*>( b ) )
                                     std::swap( a, b );
 
+                                std::lock_guard<std::mutex> lock( checkedPairsMutex );
                                 auto it = checkedPairs.find( { a, b } );
 
                                 if( it != checkedPairs.end() && it->second.layers.test( layer ) )
@@ -1100,23 +1179,8 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testGraphicClearances( )
                             // Visitor:
                             [&]( BOARD_ITEM* other ) -> bool
                             {
-                                BOARD_ITEM* a = aShape;
-                                BOARD_ITEM* b = other;
-
-                                // store canonical order so we don't collide in both directions
-                                // (a:b and b:a)
-                                if( static_cast<void*>( a ) > static_cast<void*>( b ) )
-                                    std::swap( a, b );
-
-                                auto it = checkedPairs.find( { a, b } );
-
-                                if( !testSingleLayerItemAgainstItem( aShape,
-                                                                     aShape->GetEffectiveShape().get(),
-                                                                     layer, other ) )
-                                {
-                                    if( it != checkedPairs.end() )
-                                        it->second.has_error = true;
-                                }
+                                testSingleLayerItemAgainstItem( aShape, aShape->GetEffectiveShape().get(),
+                                                                layer, other );
 
                                 return !m_drcEngine->IsCancelled();
                             },
@@ -1165,89 +1229,104 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testGraphicClearances( )
 
 void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testZonesToZones()
 {
-    bool           testClearance = !m_drcEngine->IsErrorLimitExceeded( DRCE_CLEARANCE );
-    bool           testIntersects = !m_drcEngine->IsErrorLimitExceeded( DRCE_ZONES_INTERSECT );
-    DRC_CONSTRAINT constraint;
+    bool testClearance = !m_drcEngine->IsErrorLimitExceeded( DRCE_CLEARANCE );
+    bool testIntersects = !m_drcEngine->IsErrorLimitExceeded( DRCE_ZONES_INTERSECT );
 
     std::vector<std::map<PCB_LAYER_ID, std::vector<SEG>>> poly_segments;
     poly_segments.resize( m_board->m_DRCCopperZones.size() );
 
-    // Contains the index for zoneA, zoneB, the conflict point, the actual clearance, the
-    // required clearance, and the layer
-    using report_data = std::tuple<int, int, VECTOR2I, int, int, PCB_LAYER_ID>;
+    thread_pool&        tp = GetKiCadThreadPool();
+    std::atomic<size_t> done( 0 );
+    size_t              count = 0;
 
-    std::vector<std::future<report_data>> futures;
-    thread_pool&                          tp = GetKiCadThreadPool();
-    std::atomic<size_t>                   done( 1 );
+    auto reportZoneZoneViolation = [this]( ZONE* zoneA, ZONE* zoneB, VECTOR2I& pt, int actual,
+                                           const DRC_CONSTRAINT& constraint, PCB_LAYER_ID layer ) -> void
+    {
+        std::shared_ptr<DRC_ITEM> drce;
 
-    auto checkZones =
-            [this, testClearance, testIntersects, &poly_segments, &done]
-            ( int zoneA, int zoneB, int clearance, PCB_LAYER_ID layer ) -> report_data
+        if( constraint.IsNull() )
+        {
+            drce = DRC_ITEM::Create( DRCE_ZONES_INTERSECT );
+            wxString msg = _( "(intersecting zones must have distinct priorities)" );
+            drce->SetErrorMessage( drce->GetErrorText() + wxS( " " ) + msg );
+            drce->SetItems( zoneA, zoneB );
+            reportViolation( drce, pt, layer );
+        }
+        else
+        {
+            drce = DRC_ITEM::Create( DRCE_CLEARANCE );
+            wxString msg = formatMsg( _( "(%s clearance %s; actual %s)" ), constraint.GetName(),
+                                      constraint.GetValue().Min(), std::max( actual, 0 ) );
+
+            drce->SetErrorMessage( drce->GetErrorText() + wxS( " " ) + msg );
+            drce->SetItems( zoneA, zoneB );
+            drce->SetViolatingRule( constraint.GetParentRule() );
+            ReportAndShowPathCuToCu( drce, pt, layer, zoneA, zoneB, layer, actual );
+        }
+    };
+
+    auto checkZones = [this, testClearance, testIntersects, reportZoneZoneViolation, &poly_segments,
+                       &done]( int zoneA_idx, int zoneB_idx, bool sameNet, PCB_LAYER_ID layer ) -> void
+    {
+        ZONE*    zoneA = m_board->m_DRCCopperZones[zoneA_idx];
+        ZONE*    zoneB = m_board->m_DRCCopperZones[zoneB_idx];
+        int      actual = 0;
+        VECTOR2I pt;
+
+        if( sameNet && testIntersects )
+        {
+            if( zoneA->Outline()->Collide( zoneB->Outline(), 0, &actual, &pt ) )
+            {
+                done.fetch_add( 1 );
+                reportZoneZoneViolation( zoneA, zoneB, pt, actual, DRC_CONSTRAINT(), layer );
+                return;
+            }
+        }
+        else if( !sameNet && testClearance )
+        {
+            DRC_CONSTRAINT constraint = m_drcEngine->EvalRules( CLEARANCE_CONSTRAINT, zoneA, zoneB, layer );
+            int            clearance = constraint.GetValue().Min();
+
+            if( constraint.GetSeverity() != RPT_SEVERITY_IGNORE && clearance > 0 )
             {
-                // Iterate through all the segments of refSmoothedPoly
                 std::map<VECTOR2I, int> conflictPoints;
 
-                std::vector<SEG>& refSegments = poly_segments[zoneA][layer];
-                std::vector<SEG>& testSegments = poly_segments[zoneB][layer];
-                bool reported = false;
-                auto invalid_result = std::make_tuple( -1, -1, VECTOR2I(), 0, 0, F_Cu );
+                std::vector<SEG>& refSegments = poly_segments[zoneA_idx][layer];
+                std::vector<SEG>& testSegments = poly_segments[zoneB_idx][layer];
 
+                // Iterate through all the segments in zoneA
                 for( SEG& refSegment : refSegments )
                 {
-                    int ax1 = refSegment.A.x;
-                    int ay1 = refSegment.A.y;
-                    int ax2 = refSegment.B.x;
-                    int ay2 = refSegment.B.y;
-
-                    // Iterate through all the segments in smoothed_polys[ia2]
+                    // Iterate through all the segments in zoneB
                     for( SEG& testSegment : testSegments )
                     {
-                        // Build test segment
-                        VECTOR2I pt;
-
-                        int bx1 = testSegment.A.x;
-                        int by1 = testSegment.A.y;
-                        int bx2 = testSegment.B.x;
-                        int by2 = testSegment.B.y;
-
-                        // We have ensured that the 'A' segment starts before the 'B' segment,
-                        // so if the 'A' segment ends before the 'B' segment starts, we can skip
-                        // to the next 'A'
-                        if( ax2 < bx1 )
+                        // We have ensured that the 'A' segment starts before the 'B' segment, so if the
+                        // 'A' segment ends before the 'B' segment starts, we can skip to the next 'A'
+                        if( refSegment.B.x < testSegment.A.x )
                             break;
 
-                        int d = GetClearanceBetweenSegments( bx1, by1, bx2, by2, 0,
-                                                             ax1, ay1, ax2, ay2, 0,
-                                                             clearance, &pt.x, &pt.y );
+                        int64_t  dist_sq = 0;
+                        VECTOR2I other_pt;
+                        refSegment.NearestPoints( testSegment, pt, other_pt, dist_sq );
+                        actual = std::floor( std::sqrt( dist_sq ) + 0.5 );
 
-                        if( d < clearance )
+                        if( actual < clearance )
                         {
-                            if( d == 0 && testIntersects )
-                                reported = true;
-                            else if( testClearance )
-                                reported = true;
-
-                            if( reported )
-                            {
-                                done.fetch_add( 1 );
-                                return std::make_tuple( zoneA, zoneB, pt, d, clearance, layer );
-                            }
+                            done.fetch_add( 1 );
+                            reportZoneZoneViolation( zoneA, zoneB, pt, actual, constraint, layer );
+                            return;
                         }
-
-                        if( m_drcEngine->IsCancelled() )
-                            return invalid_result;
                     }
                 }
+            }
+        }
 
-                done.fetch_add( 1 );
-                return invalid_result;
-            };
+        done.fetch_add( 1 );
+    };
 
-    for( int layer_id = F_Cu; layer_id <= B_Cu; ++layer_id )
+
+    for( PCB_LAYER_ID layer : LAYER_RANGE( F_Cu, B_Cu, m_board->GetCopperLayerCount() ) )
     {
-        PCB_LAYER_ID layer = static_cast<PCB_LAYER_ID>( layer_id );
-        int          zone2zoneClearance;
-
         // Skip over layers not used on the current board
         if( !m_board->IsLayerEnabled( layer ) )
             continue;
@@ -1256,28 +1335,26 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testZonesToZones()
         {
             if( m_board->m_DRCCopperZones[ii]->IsOnLayer( layer ) )
             {
-                SHAPE_POLY_SET poly = *m_board->m_DRCCopperZones[ii]->GetFilledPolysList( layer );
-                std::vector<SEG>& zone_layer_poly_segs = poly_segments[ii][layer];
-
-                poly.BuildBBoxCaches();
-                zone_layer_poly_segs.reserve( poly.FullPointCount() );
-
-                for( auto it = poly.IterateSegmentsWithHoles(); it; it++ )
+                if( SHAPE_POLY_SET* poly = m_board->m_DRCCopperZones[ii]->GetFill( layer ) )
                 {
-                    SEG seg = *it;
+                    std::vector<SEG>& zone_layer_poly_segs = poly_segments[ii][layer];
+                    zone_layer_poly_segs.reserve( poly->FullPointCount() );
 
-                    if( seg.A.x > seg.B.x )
-                        seg.Reverse();
+                    for( auto it = poly->IterateSegmentsWithHoles(); it; it++ )
+                    {
+                        SEG seg = *it;
 
-                    zone_layer_poly_segs.push_back( seg );
+                        if( seg.A.x > seg.B.x )
+                            seg.Reverse();
+
+                        zone_layer_poly_segs.push_back( seg );
+                    }
+
+                    std::sort( zone_layer_poly_segs.begin(), zone_layer_poly_segs.end() );
                 }
-
-                std::sort( zone_layer_poly_segs.begin(), zone_layer_poly_segs.end() );
             }
         }
 
-        std::vector<std::pair<int, int>> zonePairs;
-
         for( size_t ia = 0; ia < m_board->m_DRCCopperZones.size(); ia++ )
         {
             ZONE* zoneA = m_board->m_DRCCopperZones[ia];
@@ -1293,8 +1370,9 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testZonesToZones()
                 if( !zoneB->IsOnLayer( layer ) )
                     continue;
 
-                // Test for same net
-                if( zoneA->GetNetCode() == zoneB->GetNetCode() && zoneA->GetNetCode() >= 0 )
+                bool sameNet = zoneA->GetNetCode() == zoneB->GetNetCode() && zoneA->GetNetCode() >= 0;
+
+                if( sameNet && zoneA->GetAssignedPriority() != zoneB->GetAssignedPriority() )
                     continue;
 
                 // rule areas may overlap at will
@@ -1302,90 +1380,41 @@ void DRC_TEST_PROVIDER_COPPER_CLEARANCE::testZonesToZones()
                     continue;
 
                 // Examine a candidate zone: compare zoneB to zoneA
-                SHAPE_POLY_SET* polyA = m_board->m_DRCCopperZones[ia]->GetFill( layer );
-                SHAPE_POLY_SET* polyB = m_board->m_DRCCopperZones[ia2]->GetFill( layer );
+                SHAPE_POLY_SET* polyA = nullptr;
+                SHAPE_POLY_SET* polyB = nullptr;
+
+                if( sameNet )
+                {
+                    polyA = zoneA->Outline();
+                    polyB = zoneB->Outline();
+                }
+                else
+                {
+                    polyA = zoneA->GetFill( layer );
+                    polyB = zoneB->GetFill( layer );
+                }
 
                 if( !polyA->BBoxFromCaches().Intersects( polyB->BBoxFromCaches() ) )
                     continue;
 
-                // Get clearance used in zone to zone test.
-                constraint = m_drcEngine->EvalRules( CLEARANCE_CONSTRAINT, zoneA, zoneB, layer );
-                zone2zoneClearance = constraint.GetValue().Min();
-
-                if( constraint.GetSeverity() == RPT_SEVERITY_IGNORE || zone2zoneClearance <= 0 )
-                    continue;
-
-                futures.push_back( tp.submit( checkZones, ia, ia2, zone2zoneClearance, layer ) );
+                count++;
+                tp.push_task( checkZones, ia, ia2, sameNet, layer );
             }
         }
     }
 
-    size_t count = futures.size();
-
-    for( auto& task : futures )
+    while( true )
     {
-        if( !task.valid() )
-            continue;
+        reportProgress( done, count );
 
-        std::future_status result;
+        if( m_drcEngine->IsCancelled() )
+            break;
 
-        while( true )
-        {
-            result = task.wait_for( std::chrono::milliseconds( 250 ) );
-
-            reportProgress( done, count );
-
-            if( m_drcEngine->IsCancelled() )
-                break;
-
-            if( result == std::future_status::ready )
-            {
-                report_data  data = task.get();
-                int          zoneA_idx = std::get<0>( data );
-                int          zoneB_idx = std::get<1>( data );
-                VECTOR2I     pt = std::get<2>( data );
-                int          actual = std::get<3>( data );
-                int          required = std::get<4>( data );
-                PCB_LAYER_ID layer = std::get<5>( data );
-
-                if( zoneA_idx >= 0 )
-                {
-                    ZONE* zoneA = m_board->m_DRCCopperZones[zoneA_idx];
-                    ZONE* zoneB = m_board->m_DRCCopperZones[zoneB_idx];
-
-                    constraint = m_drcEngine->EvalRules( CLEARANCE_CONSTRAINT, zoneA, zoneB, layer );
-                    std::shared_ptr<DRC_ITEM> drce;
-
-                    if( actual <= 0 && testIntersects )
-                    {
-                        drce = DRC_ITEM::Create( DRCE_ZONES_INTERSECT );
-                    }
-                    else if( testClearance )
-                    {
-                        drce = DRC_ITEM::Create( DRCE_CLEARANCE );
-                        wxString msg = formatMsg( _( "(%s clearance %s; actual %s)" ),
-                                                  constraint.GetName(),
-                                                  required,
-                                                  std::max( actual, 0 ) );
-
-                        drce->SetErrorMessage( drce->GetErrorText() + wxS( " " ) + msg );
-                    }
-
-                    if( drce )
-                    {
-                        drce->SetItems( zoneA, zoneB );
-                        drce->SetViolatingRule( constraint.GetParentRule() );
-
-                        reportViolation( drce, pt, layer );
-                    }
-                }
-
-                break;
-            }
-        }
+        if( tp.wait_for_tasks_duration( std::chrono::milliseconds( 250 ) ) )
+            break;
     }
-}
 
+}
 
 namespace detail
 {
